"""
=================================================
Animated image using a precomputed list of images
=================================================

"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
plt.rcParams.update({'font.size': 16})
plt.rc('figure', figsize = (12, 8))
plt.rc('font', family='sans-serif', serif='Arial')
plt.rc('axes', grid = True)
plt.rc('text', usetex=True)
plt.rc('xtick', labelsize=16)
plt.rc('ytick', labelsize=16)
plt.rc('axes', labelsize=16)

def posteriorPlot(ax,figPosterior,Posterior, MAP, parNames):
    
    # Initialization
    lw = 3.
    BoundTuples = [(-5,5), (-5,5)]
    NofPa = len(MAP)
    
    # This is the true mean of the second mode that we used above:
    value1 = MAP
    
    # This is the empirical mean of the sample:
    value2 = np.mean(Posterior, axis=0)
    
    if NofPa == 2:

        #figPosterior, ax = plt.subplots()
        plt.hist2d(Posterior[:,0], Posterior[:,1], bins=(200, 200),
                   range=np.array([BoundTuples[0],BoundTuples[1]]), cmap=plt.cm.jet)
    
        plt.xlabel(parNames[0])
        plt.ylabel(parNames[1])
        
        plt.xlim(BoundTuples[0])
        plt.ylim(BoundTuples[1])
        
        ax.axvline(value1[0], color="g", lw=lw)
        ax.axhline(value1[1], color="g", lw=lw)
        ax.plot(value1[0], value1[1], "sg", lw=lw+1)
        
        #ax.axvline(value2[0], ls='--', color="r", lw=lw)
        #ax.axhline(value2[1], ls='--', color="r", lw=lw)
        #ax.plot(value2[0], value2[1], "sr", lw=lw+1)
        
    else:
        import corner
        figPosterior = corner.corner(Posterior, labels=parNames,
                           show_titles=True, title_kwargs={"fontsize": 12})
    
        # Extract the axes
        axes = np.array(figPosterior.axes).reshape((NofPa, NofPa))
        
        # Loop over the diagonal
        for i in range(NofPa):
            ax = axes[i, i]
            ax.axvline(value1[i], color="g")
            ax.axvline(value2[i], ls='--', color="r")
        
        # Loop over the histograms
        for yi in range(NofPa):
            for xi in range(yi):
                ax = axes[yi, xi]
                ax.axvline(value1[xi], color="g")
                ax.axvline(value2[xi], ls='--', color="r")
                ax.axhline(value1[yi], color="g")
                ax.axhline(value2[yi], ls='--', color="r")
                ax.plot(value1[xi], value1[yi], "sg")
                ax.plot(value2[xi], value2[yi], "sr")
    
    plt.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.1)
    #figPosterior.set_size_inches((10,10)) 
    return figPosterior

fig, ax = plt.subplots()
dirPath = '/home/farid/scientific/bayesian-validation/BayesValidRox/tests/AnalyticalFunction/Outputs_SeqPosteriorComparison/'
            
nRuns= 35
Posteriors = {}
nameList = ['init'] + list(range(1,nRuns+1))
for i, name in enumerate(nameList):
#    x = np.random.normal(size=50000)
#    y = x * 3 * (i+1) + np.random.normal(size=50000)
#    Posteriors[i] = np.stack((x,y)).T
    Posteriors[i] = np.load(dirPath+'SeqPosterior_%s.npy'%name)


# animation function
def animate(i, Posteriors): 
    
    Posterior = Posteriors[i]
    
    cont = posteriorPlot(ax,fig,Posterior, MAP = (0,0), parNames=['$X_1$', '$X_2$'])

    plt.title(r'Iteration = %i' % i)

    return cont  

anim = animation.FuncAnimation(fig, animate, fargs=(Posteriors,))#, frames=nRuns)
anim.save('animation.mp4')

     
        
#fig = plt.figure()
#
#
#def f(x, y):
#    return np.sin(x) + np.cos(y)
#
#x = np.linspace(0, 2 * np.pi, 120)
#y = np.linspace(0, 2 * np.pi, 100).reshape(-1, 1)
## ims is a list of lists, each row is a list of artists to draw in the
## current frame; here we are just animating one artist, the image, in
## each frame
#ims = []
#for i in range(60):
#    x += np.pi / 15.
#    y += np.pi / 20.
#    im = plt.imshow(f(x, y), animated=True)
#    ims.append([im])
#
#ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True,
#                                repeat_delay=1000)
#
## To save the animation, use e.g.
##
#ani.save("movie.mp4")
#
# or
#
# from matplotlib.animation import FFMpegWriter
# writer = FFMpegWriter(fps=15, metadata=dict(artist='Me'), bitrate=1800)
# ani.save("movie.mp4", writer=writer)

#plt.show()
